/*
 * Assignment2.java
 * Task:  Input Validation using regular expression
 *
 * By:  Ron Ritchey
 * Oct 7, 2007
 *
 * Note:  requires JCE Unlimited Strength Jurisdiction Policy Files (
 */


import java.io.*;
import java.util.Collections;
import java.util.TreeSet;
import java.util.TreeMap;
import java.util.Iterator;
import java.util.Map;
import java.util.logging.*;
import java.util.regex.*;
import javax.crypto.*;
import javax.crypto.spec.*;
import java.security.*;
import java.security.spec.InvalidKeySpecException;


// Class to combine the Person and Telephone objects into a single directory entry
// Allows more than one telephone number per person
class DirectoryEntry implements Comparable, Serializable {
	Person person;
	TreeMap<Telephone,Telephone> telList = new TreeMap<Telephone,Telephone>();

 	private static Logger logger =
	  Logger.getLogger(Assignment2.class.getName());

	public DirectoryEntry(Person p, Telephone t) throws InvalidPersonException, InvalidTelephoneException {
		person = new Person(p);
		Telephone nt = new Telephone(t);
		telList.put(nt, nt);
	}

	public DirectoryEntry(String name, String tel) throws InvalidPersonException, InvalidTelephoneException {
		person = new Person(name);
		Telephone nt = new Telephone(tel);
		telList.put(nt, nt);
	}

	public void print() {
		Iterator iterator = telList.entrySet().iterator();

		if (iterator.hasNext()) {
			Map.Entry e = (Map.Entry)iterator.next();
			Object o = e.getValue();
			System.out.println(person.getName() + " " + ((Telephone)o).getNumber());
		}
		while (iterator.hasNext()) {
			Map.Entry e = (Map.Entry)iterator.next();
			Object o = e.getValue();
			for (int i = 0; i < person.getName().length(); i++) System.out.print(" ");
			System.out.print(" ");
			System.out.println( ((Telephone)o).getNumber());
		}
	}

	public int compareTo(Object anotherDE) throws ClassCastException {
		if (!(anotherDE instanceof DirectoryEntry))
			throw new ClassCastException("Directory object expected");

		return person.compareTo( ((DirectoryEntry) anotherDE).person );
	}

	public final void addNumber(String number) throws InvalidTelephoneException {
		Telephone t = new Telephone(number);
		telList.put(t,t);
	}

	public final boolean removeNumber(String number) {
		Telephone t;

		try { t = new Telephone(number); } catch(InvalidTelephoneException e) { return false; }

		if (telList.remove(t) == null) return false;
		return true;

	}

	public final boolean hasNumber(String number) {
		Telephone t;
		try { t = new Telephone(number); } catch(InvalidTelephoneException e) { return false; }
		return telList.containsKey(t);
	}

	public final Person getPerson() { return person; }
	public final Telephone getTelephone() { return telList.firstKey(); }
	public final String getName() { return new String(person.getName()); }
	public final String getNumber() { return new String(telList.firstKey().getNumber()); }
	public final String[] getNumbers() {
		if (telList.size() <= 0) return null;

		String[] numbers = new String[telList.size()];

		Iterator iterator = telList.entrySet().iterator();

		int i = 0;
		while (iterator.hasNext()) {
			Map.Entry e = (Map.Entry)iterator.next();
			Object o = e.getValue();

			numbers[i] = new String(((Telephone)o).getNumber());
			i++;
		}

		return numbers;
	}

	public final int getNumbersCount() { return telList.size(); }

}


// Main class for project
public class Assignment2 {

	static TreeSet<DirectoryEntry> tsDE;

 	private static Logger logger =
	  Logger.getLogger(Assignment2.class.getName());

	private static char[] readLimited(BufferedReader in, int limit) throws IOException {
		char buf[] = new char[limit+1];

		in.read(buf, 0, limit);
		return buf;
	}

	private static char Menu() throws InvalidInputException, IOException {
		int choice;

		System.out.println("");
		System.out.println("(A)dd entry");
		System.out.println("(D)el entry");
		System.out.println("(L)ist entries");
		System.out.println("e(X)it");
		System.out.println();
		System.out.print("Enter choice: ");

		BufferedReader in = new BufferedReader(new InputStreamReader(System.in));


		String s = new String(readLimited(in, 100));
		char c = s.charAt(0);


		// Make sure that users provide a valid command
		char [] validCmds = { 'A', 'a', 'D', 'd', 'X', 'x', 'L', 'l' };

		for (int i=0; i < validCmds.length; i++) {
			if (c == validCmds[i]) return c;
		}
		throw new InvalidInputException("Bad value provided to menu");

	}

	private static boolean Add() throws InvalidInputException, InvalidPersonException, InvalidTelephoneException {
		BufferedReader in = new BufferedReader(new InputStreamReader(System.in));


		char buf[]= new char[101];
		String sName=null;
		String sNumber=null;

        	try {
			System.out.println("Please provide the following values");
			System.out.print("Name: ");
			sName = new String(readLimited(in, 100));

			System.out.print("Number: ");
			sNumber = new String(readLimited(in, 100));
		} catch(IOException e) {
			System.err.println("A system error has occured.  Please inform the administrator. Exiting...");
			logger.log(Level.SEVERE, "IOException in Assignment2.Add reading name from System.in", e);
			return false;
		}

		sName = sName.trim();
		sNumber = sNumber.trim();

		// Create new directory entry and add it to collection.  Note that if either the name or number is
		// not acceptable to the Person and Telephone classes an InvalidPersonException and/or
		// InvalidTelephoneException will be thrown here

		DirectoryEntry de = new DirectoryEntry(sName, sNumber);
		if (!tsDE.add(de)) {
			// Name already exists
			Iterator iterator = tsDE.iterator();

			// Normalize values by instantiating
			Telephone t = new Telephone(sNumber);
			Person p = new Person(sName);

			// Find matching entry
			while (iterator.hasNext()) {
				Object e = iterator.next();

				if ( ((DirectoryEntry)e).getName().equals( p.getName() ) ) {

					// If the number is not part of the current directory entry, add it
					if ( ! ((DirectoryEntry)e).hasNumber(t.getNumber() ) ) {
						((DirectoryEntry)e).addNumber(sNumber);
						return true;
					}
				}
			}
			return false;

		} else return true;
	}

	private static void Del() throws InvalidInputException, InvalidPersonException, InvalidTelephoneException {
		BufferedReader in = new BufferedReader(new InputStreamReader(System.in));


		char buf[]= new char[101];
		String sValue=null;
		int count = 0;  // Count number of deleted entries

        try {
			System.out.println("Specify entry to delete");
			System.out.print("Value: ");
			sValue = new String(readLimited(in, 100));

		} catch(IOException e) {
			System.err.println("A system error has occured.  Please inform the administrator. Exiting...");
			logger.log(Level.SEVERE, "IOException in Assignment2.Del reading name from System.in", e);
			return;
		}

		sValue = sValue.trim();

		// Try deleting as if value was a Person
		try {
			Person p = new Person(sValue);

			Iterator iterator = tsDE.iterator();

			while (iterator.hasNext()) {
				Object e = iterator.next();

				if ( ((DirectoryEntry)e).getName().equals( p.getName() ) ) {
					System.out.println("Removing " + ((DirectoryEntry)e).getName() + " " + ((DirectoryEntry)e).getNumber() );
					iterator.remove();
					count++;
				}

			}
		} catch (InvalidPersonException e) {
			// The value wasn't a person
		}


		// Try deleting like value was a telephone number
		try {
			Telephone t = new Telephone(sValue);

			Iterator iterator = tsDE.iterator();

			while (iterator.hasNext()) {
				Object e = iterator.next();

				if ( ((DirectoryEntry)e).removeNumber(t.getNumber()) ) {
					System.out.println("Removing " + t.getNumber() + " from " + ((DirectoryEntry)e).getName() );
					count++;
				}

				if ( ((DirectoryEntry)e).getNumbersCount() == 0) {
					System.out.println("Removing " + ((DirectoryEntry)e).getName() );
					iterator.remove();
				}
			}
			return;
		} catch (InvalidTelephoneException e) {
			// The value wasn't a person or a telephone number
			throw new InvalidInputException();
		} finally {
			if (count == 0) System.out.println("Value not found in directory");
		}
	}

	private static void List() {

		Iterator iterator = tsDE.iterator();

		if (iterator.hasNext()) System.out.println("List of people current in directory");
		else System.out.println("Directory Empty");

		while (iterator.hasNext()) {
			Object e = iterator.next();
			((DirectoryEntry)e).print();
		}
	}

    public static void main(String[] arg) {

		// Directory entries will be held in a TreeSet
		tsDE = new TreeSet<DirectoryEntry>();
		char c='\0';

		// The password should really be provided by the user using a routine that excepts it in,
		// computes the key and then overwrites the password from memory.
		String password = "pass phrase";

		String filename = new String("Assignment2.ser");

		// Salt will be used to perturb the key used in all of the crypto operations
		byte[] salt = 	{ 	(byte)0x27, (byte)0xa7, (byte)0x04, (byte)0xf3,
							(byte)0x24, (byte)0xe7, (byte)0xc0, (byte)0x5f
						};
		// Iteration count for PBEParameterSpec initialization
		int count = 20;


		// Start of processing.  Attempt to read in existing directory and then start main menu
		try {
			// Create Key from password
			SecretKey symmetricKey = SecretKeyFactory.getInstance("PBEWithMD5AndDES").generateSecret
				( new PBEKeySpec( password.toCharArray()) );


			PBEParameterSpec pbeParamSpec = new PBEParameterSpec(salt, count);
			Cipher pbeCipher = Cipher.getInstance("PBEWithMD5AndTripleDES");
			pbeCipher.init(Cipher.DECRYPT_MODE, symmetricKey, pbeParamSpec );

			// Attempt to open file

			boolean exist = (new File(filename)).exists();

			if ( exist ) { } else throw new FileNotFoundException();

			FileInputStream fis = new FileInputStream(filename);
			BufferedInputStream bis = new BufferedInputStream(fis);
			CipherInputStream cis = new CipherInputStream(bis, pbeCipher);
			ObjectInputStream ois = new ObjectInputStream(cis);

			tsDE = (TreeSet<DirectoryEntry>) ois.readObject();
		} catch(FileNotFoundException e) {
			// File wasn't found.  tsDE has already been initialized so no further action needed
			System.out.println("The stored directory file was not found.  Starting a new directory.\n");
		} catch(IOException e) {
			// File read unsuccessful
			logger.log(Level.SEVERE, "There was a read error while processing the stored directory file.  Exiting...", e);
			System.out.println("An error has occured during startup.  Please inform the administrator.  Exiting...");
			return;
		}  catch(InvalidKeyException e) {
			// Bad crypto setup
			logger.log(Level.SEVERE, "There was a problem initializing the decryption system.  Exiting...", e);
			System.out.println("An error has occured during startup.  Please inform the administrator.  Exiting...");
			return;
		}  catch(InvalidKeySpecException e) {
			// Bad crypto setup
			logger.log(Level.SEVERE, "There was a problem initializing the decryption system.  Exiting...", e);
			System.out.println("An error has occured during startup.  Please inform the administrator.  Exiting...");
			return;
		} catch(NoSuchAlgorithmException e) {
			// Bad crypto setup
			logger.log(Level.SEVERE, "There was a problem initializing the decryption system.  Exiting...", e);
			System.out.println("An error has occured during startup.  Please inform the administrator.  Exiting...");
			return;
		} catch(NoSuchPaddingException e) {
			// Bad crypto setup
			logger.log(Level.SEVERE, "There was a problem initializing the decryption system.  Exiting...", e);
			System.out.println("An error has occured during startup.  Please inform the administrator.  Exiting...");
			return;
		} catch(InvalidAlgorithmParameterException e) {
			// Bad crypto setup
			logger.log(Level.SEVERE, "There was a problem initializing the decryption system.  Exiting...", e);
			System.out.println("An error has occured during startup.  Please inform the administrator.  Exiting...");
			return;
		} catch(ClassNotFoundException e) {
			logger.log(Level.SEVERE, "Read an unexpected object type from stored directory file.  Exiting...", e);
			System.out.println("An error has occured during startup.  Please inform the administrator.  Exiting...");
			return;
		}

		// Main Menu
		try {

			while (c != 'x' && c != 'X') {

				try {
					c = Menu();
					switch (c) {
						case 'A':
						case 'a': if (!Add()) { System.out.println("Add failed, possible due to a duplicate name already existing in the directory"); } break;
						case 'D':
						case 'd': Del(); break;
						case 'X':
						case 'x': break;
						case 'L':
						case 'l': List(); break;
						default: throw new InvalidInputException("Bad value provided to menu");
					}
				} catch(InvalidInputException e) {
					logger.finest("Received invalid input to menu");
				} catch(InvalidPersonException e) {
					logger.finest("Received invalid person from user during menu call");
					System.out.println("Unable to process provided name");
				} catch(InvalidTelephoneException e) {
					logger.finest("Received invalid telephone from user during menu call");
					System.out.println("Unable to process provided telephone number");
				}
			}

	    } catch(IOException e) {
			logger.log(Level.SEVERE, "IOException", e);
		} catch(Exception e) {
			logger.log(Level.SEVERE, "Other Exception", e);
		} finally {

			// Shutting down.  Attempt to save directory to disk.
			try {
				// Create Key from password
				SecretKey symmetricKey = SecretKeyFactory.getInstance("PBEWithMD5AndDES").generateSecret
					( new PBEKeySpec( password.toCharArray()) );


				PBEParameterSpec pbeParamSpec = new PBEParameterSpec(salt, count);
				Cipher pbeCipher = Cipher.getInstance("PBEWithMD5AndTripleDES");
				pbeCipher.init(Cipher.ENCRYPT_MODE, symmetricKey, pbeParamSpec );

				FileOutputStream fos = new FileOutputStream(filename);
				BufferedOutputStream bos = new BufferedOutputStream(fos);
				CipherOutputStream cos = new CipherOutputStream(bos, pbeCipher);
				ObjectOutputStream oos = new ObjectOutputStream(cos);

				oos.writeObject(tsDE);
				oos.flush();
				oos.close();
			} catch(IOException e) {
				// File write unsuccessful
				logger.log(Level.SEVERE, "There was a write error while processing the stored directory file.  Exiting...", e);
				System.out.println("An error has occured during shutdown.  Please inform the administrator.  Exiting...");
				return;
			}  catch(InvalidKeyException e) {
				// Bad crypto setup
				logger.log(Level.SEVERE, "There was a problem initializing the encryption system.  Exiting...", e);
				System.out.println("An error has occured during shutdown.  Please inform the administrator.  Exiting...");
				return;
			}  catch(InvalidKeySpecException e) {
				// Bad crypto setup
				logger.log(Level.SEVERE, "There was a problem initializing the encryption system.  Exiting...", e);
				System.out.println("An error has occured during shutdown.  Please inform the administrator.  Exiting...");
				return;
			} catch(NoSuchAlgorithmException e) {
				// Bad crypto setup
				logger.log(Level.SEVERE, "There was a problem initializing the encryption system.  Exiting...", e);
				System.out.println("An error has occured during shutdown.  Please inform the administrator.  Exiting...");
				return;
			} catch(NoSuchPaddingException e) {
				// Bad crypto setup
				logger.log(Level.SEVERE, "There was a problem initializing the encryption system.  Exiting...", e);
				System.out.println("An error has occured during shutdown.  Please inform the administrator.  Exiting...");
				return;
			} catch(InvalidAlgorithmParameterException e) {
				// Bad crypto setup
				logger.log(Level.SEVERE, "There was a problem initializing the encryption system.  Exiting...", e);
				System.out.println("An error has occured during shutdown.  Please inform the administrator.  Exiting...");
				return;
			}

		}
    }
}
