/*
 * Person.java
 * Purpose:  Provides Person storage and validation for Assignment 2
 *
 * By:  Ron Ritchey
 * Oct 7, 2007
 *
 *
 */

import java.util.logging.*;
import java.util.regex.*;
import java.io.*;

class PATTERNS {

}

public class Person implements Comparable, Serializable, ObjectInputValidation {
	String FirstName=new String("");
	String MiddleName=new String("");
	String LastName=new String("");
	String s;

	public final static String namePattern = "[a-zA-Z]+('[a-zA-Z]+)?(-[a-zA-Z]+('[a-zA-Z]+)?)?";
	public final static String FirstMidLastPattern = "^("+namePattern+" ){0,2}"+namePattern+"$";
	public final static String LastMidFirstPattern = "^"+namePattern+", ?("+namePattern+" )?"+namePattern+"$";
	public final static String FullNamePattern = "("+FirstMidLastPattern+")|("+LastMidFirstPattern+")";

	private static String validate(String input, String pattern) throws InvalidInputException {
		Pattern p = Pattern.compile(pattern);
		Matcher m = p.matcher(input);

		if (!m.find()) throw new InvalidInputException();

		return m.group();
	}

	private static Logger logger =
	  Logger.getLogger(Person.class.getName());

	public Person(Person p) throws InvalidPersonException {
		try { validate(p.getName(), FullNamePattern); }
		catch(InvalidInputException e) { logger.fine("Received invalid name"); throw new InvalidPersonException(); }

		FirstName = new String(p.FirstName);
		MiddleName = new String(p.MiddleName);
		LastName = new String(p.LastName);
	}

	public Person(String name) throws InvalidPersonException {

		// Make sure we got a good name
		try { validate(name, FullNamePattern); }
		catch(InvalidInputException e) { logger.fine("Received invalid name"); throw new InvalidPersonException(); }


		// Is this a last, first middle name?
		try {
			// Creating string that we will consume as we parse it
			s = new String(name);

			validate(s, LastMidFirstPattern);
			// No exception so this must be a last, middle first

			// Extract LastName
			LastName = validate(s, namePattern+", ?");
			s = s.substring(LastName.length());

			LastName = LastName.replaceAll("[\\s,]", "");

			// Extract FirstName
			FirstName = validate(s, namePattern+" ?");
			s = s.substring(FirstName.length());
			FirstName = FirstName.replaceAll("[\\s]", "");

			// Extract MiddleName
			MiddleName = validate(s, namePattern);
			MiddleName = MiddleName.replaceAll("[\\s]", "");
			return;
		}
		catch(InvalidInputException e) {
			// If LastName has a value then we have succeeded in getting a valid name
			if (!LastName.equals("")) return;
		}

		// Is this a first middle last name?
		try {
			// Creating string that we will consume as we parse it
			s = new String(name);

			validate(s, FirstMidLastPattern);

			// No exception so this must be a first mid last

			// Extract First Name
			FirstName = validate(s, namePattern+" ?");
			s = s.substring(FirstName.length());
			FirstName = FirstName.replaceAll("[\\s]", "");

			// Extract Middle Name
			MiddleName = validate(s, namePattern+" ?");
			s = s.substring(MiddleName.length());
			MiddleName = MiddleName.replaceAll("[\\s]", "");

			// Extract Last Name
			LastName = validate(s, namePattern);
			LastName = LastName.replaceAll("[\\s]", "");
		}
		catch(InvalidInputException e) {
			// If last name was populated then name had all three components
			if (LastName.length() > 0) return;
			else {
				if (MiddleName.length() > 0) {
					// This is two part name because last was empty, but middle had value
					LastName = MiddleName;
					MiddleName = "";
					return;
				} else {
					if (FirstName.length() > 0) {
						// This is one part name (e.g. Cher)
						LastName = FirstName;
						FirstName = "";
						return;
					} else throw new InvalidPersonException("Received bad name that passed initial pattern check");
				}
			}
		}

	} // End of Person(string name)


	public String getName() {
		if (FirstName.equals("")) return LastName; else
		if (MiddleName.equals("")) return FirstName + " " + LastName; else
		return FirstName + " " + MiddleName + " " + LastName;
	}
	public String getLastMidFirst() {
		if (MiddleName.equals("")) return LastName+", "+FirstName; else
		return LastName + ", " + FirstName + " " + MiddleName;
	}
	public String getFirstName() { return FirstName; }
	public String getMiddleName() { return MiddleName; }
	public String getLastName() { return LastName; }

	public int compareTo(Object anotherPerson) throws ClassCastException {
		if (!(anotherPerson instanceof Person))
			throw new ClassCastException("Person object expected");

		if(LastName.compareTo( ((Person) anotherPerson).getLastName() ) == 0) {
			if (FirstName.compareTo( ((Person) anotherPerson).getFirstName() ) == 0) {
				if (MiddleName.compareTo( ((Person) anotherPerson).getMiddleName()) == 0) return 0;
				else return MiddleName.compareTo( ((Person) anotherPerson).getMiddleName());
			}
			else return FirstName.compareTo( ((Person) anotherPerson).getFirstName() );
		}
		else return LastName.compareTo( ((Person) anotherPerson).getLastName() );
	}

	private void readObject(ObjectInputStream ois) throws ClassNotFoundException, IOException {
		ois.registerValidation(this, 0);
		ois.defaultReadObject();
	}

	public void validateObject() throws InvalidObjectException {
		try {
			Person p = new Person(getName());
		} catch (InvalidPersonException e) {
			throw new InvalidObjectException("Invalid Person value in stream");
		}
	}
}
